! WHIZARD 1.13 Aug 06 2001
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module analyze

  use kinds, only: default !NODEP!
!  use kinds, only: double !NODEP!
  use limits, only: MAX_EXTERNAL, CUT_MODE_STRLEN, CUT_TEX_STRLEN
  use limits, only: ANALYZE_MAX, HISTOGRAMS_MAX
  use limits, only: HELICITIES_MAX, FLAVORS_MAX
  use limits, only: DEFAULT_FILENAME, BUFFER_SIZE
  use constants
  use lorentz
  use particles
  use events
  use decay_trees, only: TC
  use file_utils !NODEP!
  use parser
!  use particles, only: particle_tex_form
  use cuts
  use histograms !NODEP!
  use monte_carlo_states
  use diagnostics, only: msg_level, msg_buffer
  use diagnostics, only: msg_message, msg_warning, msg_error

  implicit none
  private

  integer, parameter, public :: TEXT_FMT=0, GML_FMT=1

  type, public :: histogram_data
     private
     integer :: mode
     integer(kind=TC), dimension(2) :: code
     type(histogram) :: h
  end type histogram_data
  type, public :: analysis_data
     private
     integer :: n_histograms, n_evt, n_evt_total
     real(kind=default) :: sum_evt, sum_evt_sq, sum_evt_excess, max_weight
     type(cut_array) :: cc
     integer :: n_hel, n_flv
     integer, dimension(MAX_EXTERNAL, HELICITIES_MAX) :: hel
     integer, dimension(MAX_EXTERNAL, FLAVORS_MAX) :: flv
     type(histogram_data), dimension(HISTOGRAMS_MAX) :: hh
  end type analysis_data
  type, public :: analysis_bundle
     private
     integer :: n_datasets
     type(analysis_data), dimension(ANALYZE_MAX) :: a
  end type analysis_bundle

  public :: create, destroy
  public :: read, write
  public :: apply

  interface create
     module procedure analysis_bundle_create
     module procedure analysis_data_create
     module procedure histogram_data_create
  end interface
  interface destroy
     module procedure analysis_bundle_destroy
     module procedure analysis_data_destroy
     module procedure histogram_data_destroy
  end interface
  interface read
     module procedure analysis_bundle_read_name
     module procedure analysis_bundle_read_unit
     module procedure analysis_data_read_unit
  end interface
  interface write
     module procedure analysis_data_write_unit
     module procedure analysis_bundle_write_unit
     module procedure analysis_bundles_write_name
  end interface
  interface apply
     module procedure analysis_bundle_apply
     module procedure analysis_data_apply
     module procedure histogram_data_apply
  end interface

contains

  subroutine analysis_bundle_create(aa)
    type(analysis_bundle), intent(out) :: aa
    aa%n_datasets = 0
  end subroutine analysis_bundle_create
  subroutine analysis_bundle_destroy(aa)
    type(analysis_bundle), intent(inout) :: aa
    integer :: i
    do i=1, aa%n_datasets
       call destroy(aa%a(i))
    end do
    aa%n_datasets = 0
  end subroutine analysis_bundle_destroy
  
  subroutine analysis_data_create(a)
    type(analysis_data), intent(out) :: a
    call create(a%cc)
    a%n_histograms = 0
    a%sum_evt = 0
    a%sum_evt_sq = 0
    a%sum_evt_excess = 0
    a%max_weight = 0
    a%n_evt = 0
    a%n_evt_total = 0
  end subroutine analysis_data_create
  subroutine analysis_data_destroy(a)
    type(analysis_data), intent(inout) :: a
    integer :: i
    do i=1, a%n_histograms
       call destroy(a%hh(i))
    end do
    a%n_histograms = 0
    a%sum_evt = 0
    a%sum_evt_sq = 0
    a%sum_evt_excess = 0
    a%max_weight = 0
    a%n_evt = 0
    a%n_evt_total = 0
   call destroy(a%cc)
  end subroutine analysis_data_destroy

  subroutine histogram_data_create(h, mode, code, lower, upper, nbin, step)
    type(histogram_data), intent(out) :: h
    integer, intent(in) :: mode
    integer(kind=TC), dimension(2), intent(in) :: code
    real(kind=default), intent(in) :: lower, upper
    integer, intent(in), optional :: nbin
    real(kind=default), intent(in), optional :: step
    real(kind=default) :: x_min, x_max
    if (mode/=NO_CUT) then
       h%mode = mode
       h%code = code
       if (lower <= upper) then
          x_min = lower
          x_max = upper
       else
          x_min = upper
          x_max = lower
       end if
       if (present(nbin)) then
          if (nbin>0) then
             call create_histogram(h%h, x_min, x_max, nbin)
          else
             call create_histogram(h%h, x_min, x_max, 1)
          end if
       else if (present(step)) then
          if (step/=0) then
             call create_histogram(h%h, x_min, x_max, &
                  &                nint((x_max-x_min)/abs(step)))
          else
             call create_histogram(h%h, x_min, x_max, 1)
          end if
       else
          call create_histogram(h%h, x_min, x_max)
       end if
    end if
  end subroutine histogram_data_create

  subroutine histogram_data_destroy(h)
    type(histogram_data), intent(inout) :: h
    h%mode = NO_CUT
    h%code = (/ 0, 0 /)
    call delete_histogram(h%h)
  end subroutine histogram_data_destroy

  subroutine analysis_bundle_read_name(filename, aa, process_id)
    character(len=*), intent(in) :: filename
    type(analysis_bundle), intent(inout) :: aa
    character(len=*), intent(in) :: process_id
    character(len=max(len(filename), len(DEFAULT_FILENAME))+5) :: file
    integer :: unit
    logical :: exist
    file = trim(filename)//".cut5"
    inquire(file=trim(file), exist=exist)
    if (.not.exist) then
       file = trim(DEFAULT_FILENAME)//".cut5"
       inquire(file=trim(file), exist=exist)
    end if
    if (exist) then
       unit = free_unit()
       call msg_message
       call msg_message (" Reading analysis configuration data from file " &
            &            // trim(file))
       open(unit=unit, file=trim(file), status='old', action='read')
       call analysis_bundle_read_unit(unit, aa, process_id)
       if (aa%n_datasets>0) then
          write (msg_buffer, "(1x,A,1x,I3,1x,A)") &
               & "Found", aa%n_datasets, "analysis configuration datasets."
          call msg_message
       else
          call msg_message (" No analysis data found for process " // &
               &            trim(process_id))
          call analysis_bundle_default (aa)
       end if
       call lex_clear
       close(unit)
    else
       call msg_message (" Analysis configuration file " // &
            &            trim(file) // " not found.")
       call analysis_bundle_default (aa)
    end if
  end subroutine analysis_bundle_read_name

  subroutine analysis_bundle_read_unit(unit, aa, process_id)
    integer, intent(in) :: unit
    type(analysis_bundle), intent(inout) :: aa
    character(len=*), intent(in) :: process_id
    logical :: ok
    integer :: iostat, i
    FIND_PROCESS: do
       call find_string(unit, 'process', ok=ok)
       if (.not.ok) exit FIND_PROCESS
       call find_string(unit, process_id, &
            & reject=(/ 'process  ', 'grove    ', 'tree     ', &
            &           'cut      ', 'helicity ', 'flavor   ', &
            &           'histogram', 'and      ' /), &
            & ok=ok)
       if (ok) exit FIND_PROCESS
    end do FIND_PROCESS
    if (.not.ok) return
    i = 1
    SCAN_ANALYSIS_BUNDLE: do
       call create(aa%a(i))
       call analysis_data_read_unit(unit, aa%a(i))
       aa%n_datasets = i
       i = i+1
       if (i > ANALYZE_MAX) exit SCAN_ANALYSIS_BUNDLE
       call find_string(unit, 'and', here=.true., ok=ok, iostat=iostat)
       if (.not.ok) exit SCAN_ANALYSIS_BUNDLE
    end do SCAN_ANALYSIS_BUNDLE
  end subroutine analysis_bundle_read_unit

  subroutine analysis_data_read_unit(unit, a)
    integer, intent(in) :: unit
    type(analysis_data), intent(inout) :: a
    logical :: ok
    integer :: iostat
    integer :: mode
    integer(kind=TC), dimension(2) :: code
    character(len=CUT_MODE_STRLEN) :: str
    real(kind=default) :: lower, upper, step
    integer :: i, nbin
    call read(unit, a%cc, level=5)
    SCAN_HELICITIES: do
       call find_string (unit, 'helicity', here=.true., ok=ok, iostat=iostat)
       if (ok) then
          a%n_hel = a%n_hel+1
          do i=1, MAX_EXTERNAL
             a%hel(i, a%n_hel) = get_integer(unit, ok, iostat)
          end do
       else
          call find_string (unit, 'flavor', here=.true., ok=ok, iostat=iostat)
          if (.not.ok) exit SCAN_HELICITIES
          a%n_flv = a%n_flv+1
          do i=1, MAX_EXTERNAL
             a%flv(i, a%n_flv) = get_integer(unit, ok, iostat)
          end do
       end if
    end do SCAN_HELICITIES
    SCAN_ANALYZE: do
       call find_string(unit, 'histogram', here=.true., ok=ok, iostat=iostat)
       if (.not.ok) exit SCAN_ANALYZE
       i = a%n_histograms+1
       str = get_string_upper_case(unit)
       select case(trim(str))
         case("- ", "--", "---")
            mode = NO_CUT
         case("M","Q")
            mode = M_CUT
         case("LM","LQ")
            mode = LM_CUT
         case("MSQ","QSQ","S","T","U")
            mode = MSQ_CUT
         case("E")
            mode = E_CUT
         case("LE")
            mode = LE_CUT
         case("PT")
            mode = PT_CUT
         case("PL")
            mode = PL_CUT
         case("P")
            mode = P_CUT
         case("RAP", "RAPIDITY", "Y")
            mode = RAP_CUT
         case("ETA")
            mode = ETA_CUT
         case("A", "ANGLE", "TH", "THETA")
            mode = A_CUT
         case("CT","COS(TH)","COS(THETA)")
            mode = CT_CUT
         case("AD", "ANGLE(DEG)", "TH(DEG)", "THETA(DEG)")
            mode = AD_CUT
         case("PH", "PHI")
            mode = PHI_CUT
         case("PHD", "PHID", "PHI(DEG)")
            mode = PHID_CUT
         case("DETA", "DELTA-ETA")
            mode = DETA_CUT
       case default
          call msg_message (" histogram "//trim(str)//" ...")
          call msg_error (" This is not a valid histogram specifier.  I'll ignore this.")
          mode = NO_CUT
       end select
       call find_string(unit, 'of', here=.true.)
       code(1) = get_integer(unit)
       code(2) = get_integer(unit, ok)
       if (.not.ok) code(2) = 0
       call find_string(unit, 'within', here=.true.)
       lower = get_real(unit)
       upper = get_real(unit)
       call find_string(unit, 'nbin', here=.true., ok=ok)
       if (ok) then
          nbin = get_integer(unit)
          if (mode/=NO_CUT) &
               & call create(a%hh(i), mode, code, lower, upper, nbin=nbin)
       else
          call find_string(unit, 'step', here=.true., ok=ok)
          if (ok) then
             step = get_real(unit)
             if (mode/=NO_CUT) &
                  & call create(a%hh(i), mode, code, lower, upper, step=step)
          else
             if (mode/=NO_CUT) call create(a%hh(i), mode, code, lower, upper)
          end if
       end if
       a%n_histograms = i
    end do SCAN_ANALYZE
  end subroutine analysis_data_read_unit

  subroutine analysis_bundle_default (aa)
    type(analysis_bundle), intent(inout) :: aa
    call create(aa%a(1))
    aa%n_datasets = 1
  end subroutine analysis_bundle_default

  subroutine analysis_bundles_write_name (name, fmt, aa, mcs)
    character(len=*), intent(in) :: name
    integer, intent(in) :: fmt
    type(analysis_bundle), dimension(:), intent(in) :: aa
    type(monte_carlo_state), dimension(:), intent(in) :: mcs
    integer :: u
    character(len=BUFFER_SIZE) :: fname
    u = free_unit()
    select case (fmt)
    case (TEXT_FMT)
       do mc_process_index=1, mc_number_processes
          fname = trim(name) // "." &
               & // trim(mcs(mc_process_index)%process_id_current) // ".dat"
          open (unit=u, action='write', status='replace', file=fname)
          call analysis_bundle_write_unit &
               & (u, aa(mc_process_index), mcs(mc_process_index))
          close (u)
       end do
    case(GML_FMT)
       fname = trim(name) // "-plots.tex"
       open (unit=u, action='write', status='replace', file=fname)
       call gml_bundles_write_unit (u, aa, mcs, trim(name))
       close (u)
    end select
  end subroutine analysis_bundles_write_name

  subroutine analysis_bundle_write_unit (u, aa, mcs, show_histograms)
    integer, intent(in) :: u
    type(analysis_bundle), intent(in) :: aa
    type(monte_carlo_state), intent(in) :: mcs
    logical, intent(in), optional :: show_histograms
    integer :: i
    do i=1, aa%n_datasets
       call write(u, trim(mcs%process_id_current), aa%a(i), &
            &     mcs%it, mcs%integral, mcs%n_in, mcs%n_out, &
            &     mcs%code(:,1), &
            &     show_histograms=show_histograms, &
            &     show_overflow=mcs%show_overflow, &
            &     show_excess=mcs%show_excess.and.mcs%unweighted)
    end do
  end subroutine analysis_bundle_write_unit

  subroutine analysis_data_write_unit &
       & (u, process_id, a, it, integral, n_in, n_out, code, &
       &  show_histograms, show_overflow, show_excess)
    integer, intent(in) :: u
    character(len=*), intent(in) :: process_id
    type(analysis_data), intent(in) :: a
    integer, intent(in) :: it
    real(kind=default), intent(in) :: integral
    integer, intent(in) :: n_in, n_out
    integer, intent(in), dimension(:) :: code
    logical, intent(in), optional :: &
         & show_histograms, show_overflow, show_excess
    real(kind=default) :: eff, integral_cut, error_cut, accuracy
    logical :: show_hist
    integer :: i
    call mc_write_dline(u)
    if (u /= 6) then
       write(u,*) '! WHIZARD 1.13 (Aug 06 2001)'
       call write_process (u, process_id, code, n_in, n_out)
    end if
    if (u /= 6 .or. msg_level > 3)  call msg_message &
         & (" Analysis results for process "//trim(process_id)//":", unit=u)
    if (cut_number(a%cc)>0) then
       if (u /= 6 .or. msg_level > 3) then
          call msg_message (unit=u)
          call msg_message (" Additional cuts:", unit=u)
          call write(u, a%cc)
       end if
    end if
    if (a%n_hel>0) then
       if (u /= 6 .or. msg_level > 3) then
          call msg_message (unit=u)
          call msg_message (" Allowed helicity combinations:", unit=u)
          do i=1, a%n_hel
             write (msg_buffer, "(12(1x,I2))") a%hel(:n_out,i)
             call msg_message (unit=u)
          end do
       end if
    end if
    if (a%n_flv>0) then
       if (u /= 6 .or. msg_level > 3) then
          call msg_message (unit=u)
          call msg_message (" Allowed flavor combinations:", unit=u)
          do i=1, a%n_flv
             write (msg_buffer, "(12(1x,I3))") a%flv(:n_out,i)
             call msg_message (unit=u)
          end do
       end if
    end if
    show_hist = .true.
    if (present(show_histograms)) show_hist = show_histograms
    if (show_hist .and. a%n_histograms>0) then
       call msg_message (unit=u)
       call msg_message (" Histograms:", unit=u)
       do i=1, a%n_histograms
          call msg_message (unit=u)
          call histogram_data_write_unit &
               & (u, a%hh(i), show_overflow, show_excess)
       end do
    end if
    if (a%n_evt_total==0) return
    call mc_write_header(u)
    eff = a%sum_evt / real(a%n_evt_total, kind=default)
    integral_cut = a%sum_evt * integral / real(a%n_evt_total, kind=default)
    error_cut = &
         & sqrt(a%sum_evt_sq) * integral / real(a%n_evt_total, kind=default)
    accuracy = mc_get_accuracy (integral_cut, error_cut, a%n_evt)
    call mc_write_result &
         & (u, it, a%n_evt, integral_cut, error_cut, eff, accuracy)
    if (a%sum_evt_excess>0) then
       call mc_write_hline(u)
       write(msg_buffer, "(1x,A,1x,F6.1,7x,A,1x,2PF7.2,A,2x,A,0PF6.2)") &
            & "Excess events:", a%sum_evt_excess, &
            & "(", a%sum_evt_excess / a%sum_evt, "% )", &
            & "| Maximal weight:", a%max_weight
       call msg_warning (unit=u)
    end if
  end subroutine analysis_data_write_unit

  subroutine histogram_data_write_unit(u, h, show_overflow, show_excess)
    integer, intent(in) :: u
    type(histogram_data), intent(in) :: h
    logical, intent(in), optional :: show_overflow, show_excess
    character(len=CUT_MODE_STRLEN) :: str
    select case(h%mode)
    case(M_CUT);  str = "M"
    case(LM_CUT);  str = "LM"
    case(MSQ_CUT);  str = "MSQ"
    case(E_CUT);  str = "E "
    case(LE_CUT);  str = "LE"
    case(PT_CUT);  str = "pT"
    case(PL_CUT);  str = "pL"
    case(P_CUT);  str = "p "
    case(RAP_CUT);  str = "y"
    case(ETA_CUT);  str = "eta"
    case(A_CUT);  str = "theta"
    case(CT_CUT);  str = "cos(theta)"
    case(AD_CUT);  str = "theta(DEG)"
    case(PHI_CUT);  str = "phi"
    case(PHID_CUT);  str = "phi(DEG)"
    case(DETA_CUT);  str = "delta-eta"
    end select
    if (h%mode /= NO_CUT) then
       write(u, '(1x,A,1x,A,1x,A,1x,I3)', advance='no') &
            & 'histogram', trim(str), 'of', h%code(1)
       if (h%code(2)/=0) then
          write(u, '(1x,I3)', advance='no') h%code(2)
       else
          write(u, '(1x,3x)', advance='no')
       end if
       write(u, '(1x,A)', advance='no') 'within'
       write(u, '(2(1x,1PE12.5))', advance='no') h%h%x_min, h%h%x_max
       write(u, '(2x,A,1x,I4)') 'nbin', h%h%n_bins
       call write_histogram(h%h, u, show_overflow, show_excess)
    end if
  end subroutine histogram_data_write_unit

  subroutine gml_bundles_write_unit (u, aa, mcs, name)
    integer, intent(in) :: u
    type(analysis_bundle), dimension(:), intent(in) :: aa
    type(monte_carlo_state), dimension(:), intent(in) :: mcs
    character(len=*), intent(in) :: name
    integer :: i, j
    write(u, '(A)') "\documentclass[12pt]{article}"
    write(u, *)
    write(u, '(A)') "\usepackage{gamelan}"
    write(u, '(A)') "\usepackage{amsmath}"
    write(u, *)
    write(u, '(A)') "\begin{document}"
    write(u, '(A)') "\begin{gmlfile}"
    write(u, *)
    do j=1, size (aa)
       do i=1, aa(j)%n_datasets
          if (i>1 .or. j>1) write(u, '(A)') "\newpage"
          write(u, '(A)') repeat("%", 72)
          call gml_data_write_unit &
               & (u, aa(j)%a(i), mcs(j), name)
       end do
    end do
    write(u, *)
    write(u, '(A)') "\end{gmlfile}"
    write(u, '(A)') "\end{document}"
  end subroutine gml_bundles_write_unit

  subroutine gml_data_write_unit (u, a, mcs, name)
    integer, intent(in) :: u
    type(analysis_data), intent(in) :: a
    type(monte_carlo_state), intent(in) :: mcs
    character(len=*), intent(in) :: name
    real(kind=default) :: eff, integral_cut, error_cut
    real(kind=default) :: rel_error, rel_error_cut, accuracy
    character(len=32) :: dx_string, x_low_string, x_high_string
    character(len=CUT_TEX_STRLEN) :: str, unit
    integer :: i
    write(u, '(A)') "\noindent\textbf{\large\texttt{WHIZARD} data analysis}" &
         & // "\hfill\today"
    write(u, '(A)') "\begin{flushleft}"
    write(u, '(A)') "\textbf{Process:} "
    write(u, '(2x,A)', advance='no') "\verb|" &
         & // trim(mcs%process_id_current) // "| ("
    write(u, '(A)', advance='no') "$"
    do i=1, mcs%n_tot
       write(u, '(1x,A)', advance='no') trim(particle_tex_form(mcs%code(i,1)))
       if (i==mcs%n_in) write(u, '(1x,A)', advance='no') "\to"
    end do
    write(u, '(A)') "$) \\"
    if (a%n_hel>0) then
       write(u, '(A)') "\textbf{Allowed helicity combinations:} "
       write(u, '(A)') "$"
       do i=1, a%n_hel
          write(u, *) "(", a%hel(:mcs%n_out,i), ")\ "
       end do
       write(u, '(A)') "$ \\"
    end if
    if (a%n_flv>0) then
       write(u, '(A)') "\textbf{Allowed flavor combinations:} "
       write(u, '(A)') "$"
       do i=1, a%n_flv
          write(u, *) "(", a%flv(:mcs%n_out,i), ")\ "
       end do
       write(u, '(A)') "$ \\"
    end if
    write(u, '(A)') "\end{flushleft}"
    write(u, '(A)') "%"
    if (cut_number(a%cc)>0) then
       write(u, '(A)') "\begin{center}"
       call tex_write(u, a%cc)
       write(u, '(A)') "\end{center}"
       write(u, '(A)') "%"
    end if
    if (mcs%n_in==2) then
       write(u, '(A)') '\begin{displaymath}'
       write(u, 10) '\sqrt{s} = ', mcs%sqrts, '\;\textrm{GeV}\qquad'
       write(u, 10) '\int\mathcal{L} = ', &
            & mcs%luminosity, '\;\textrm{fb}^{-1}'
       write(u, '(A)') '\end{displaymath}'
    else
       write(u, '(A)') '\begin{displaymath}'
       write(u, 10) 'M = ', mcs%sqrts, '\;\textrm{GeV}\\'
       write(u, '(A)') '\end{displaymath}'
    end if
    if (a%n_histograms>0) then
       write(u, '(A)') "\begin{center} \unitlength 1mm"
       do i=1, a%n_histograms
          unit = ""
          select case(a%hh(i)%mode)
          case(M_CUT)
             str  = "$M_{\rm inv}$"
             unit = "[\textrm{GeV}]"
          case(LM_CUT)
             str  = "$\log_{10}M_{\rm inv}/\textrm{GeV}$"
          case(MSQ_CUT)
             str  = "$M^2_{\rm inv}$"
             unit = "[$\textrm{GeV}^2$]"
          case(E_CUT)
             str  = "$E$"
             unit = "[\textrm{GeV}]"
          case(LE_CUT)
             str  = "$\log_{10}E/\textrm{GeV}$"
          case(PT_CUT)
             str  = "$p_\perp$"
             unit = "[\textrm{GeV}]"
          case(PL_CUT)
             str = "$p_\parallel$"
             unit = "[\textrm{GeV}]"
          case(P_CUT)
             str  = "$|\vec p\,|$"
             unit = "[\textrm{GeV}]"
          case(RAP_CUT)
             str  = "$y$"
             unit = "(rap.)"
          case(ETA_CUT) 
                str  = "$\eta$"
          case(A_CUT)
             str  = "$\theta$"
             unit = "[\textrm{rad}]"
          case(CT_CUT)
             str  = "$\cos\theta$"
          case(AD_CUT)
             str  = "$\theta$"
             unit = "[\textrm{deg.}]"
          case(PHI_CUT)
             str  = "$\phi$"
             unit = "[\textrm{rad}]"
          case(PHID_CUT)
             str  = "$\phi$"
             unit = "[\textrm{deg.}]"
          case(DETA_CUT) 
                str  = "$\Delta\eta$"
          end select
          dx_string = " "; x_low_string = " ";  x_high_string = " "
          write(x_low_string, '(G10.3)') a%hh(i)%h%x_min
          x_low_string = adjustl(x_low_string)
          write(x_high_string, '(G10.3)') a%hh(i)%h%x_max
          x_high_string = adjustl(x_high_string)
          write(dx_string, '(G10.3)') &
               & (a%hh(i)%h%x_max - a%hh(i)%h%x_min) / (2*a%hh(i)%h%n_bins)
          dx_string = adjustl(dx_string)
          write(u, '(A)', advance='no') '\begin{gmlgraph*}('
          write(u, *) mcs%plot_width, ', ', mcs%plot_height, ')[dat]'
          write(u, '(2x,A)') '%setup(linear,log); ' &
               & //'graphrange (#'//trim(x_low_string) &
               & //',??), (#'//trim(x_high_string)//',??);'
          write(u, '(2x,A)') 'setup(linear,linear); ' &
               & //'graphrange (#'//trim(x_low_string) &
               & //',#0), (#'//trim(x_high_string)//',??);'
          if (mcs%show_overflow) then
             write(u, '(2x,A)') 'endkey "overflow";'
          else
             write(u, '(2x,A)') 'endkey "!";'
          end if
          write(u, '(2x,A)') 'fromfile "' &
               & // name // "." // trim(mcs%process_id_current) // '.dat":' 
          if (mcs%show_overflow) then
             write(u, '(4x,A)') 'key "        underflow";'
          else
             write(u, '(4x,A)') 'key " histogram";'
          end if
          write(u, '(4x,A)') 'dx := #'//trim(dx_string)//';'
          write(u, '(4x,A)') 'for i withinblock:' 
          write(u, '(6x,A)') 'get x,y;'
          write(u, '(6x,A)') 'plot(dat)(x,y) hbar dx;'
          write(u, '(4x,A)') 'endfor'
          write(u, '(4x,A)') 'calculate dat.base(dat)(x,#0);'
          write(u, '(2x,A)') 'endfrom'
          write(u, '(2x,A)') 'fill piecewise from(dat, dat.base/\) '// &
               & 'withcolor .9white outlined;'
          write(u, '(2x,A)') 'label.ulft(<'//'<\#evt/bin>'//'>, out);'
          write(u, '(2x,A)', advance='no') 'label.bot(<'//'<'
          if (a%hh(i)%code(2)==0) then
             write(u, '(1x,A,1x,A,1x,A,1x,I3,A)', advance='no') &
                  & trim(str), trim(unit), 'of $(', a%hh(i)%code(1), ')$'
          else
             write(u, '(1x,A,1x,A,1x,A,1x,I3,A,I3,A)', advance='no') &
                  & trim(str), trim(unit), 'of $(', a%hh(i)%code(1), ', ', &
                  & a%hh(i)%code(2), ')$'
          end if
          write(u, '(A)') '>'//'>, out);'
          write(u, '(A)') '\end{gmlgraph*}'
       end do
       write(u, '(A)') "\end{center}"
    end if
    if (mcs%integral/=0) then
       rel_error = mcs%error/mcs%integral
    else
       rel_error = 0
    end if
    if (a%n_evt_total/=0) then
       eff = a%sum_evt / real(a%n_evt_total, kind=default)
       integral_cut = a%sum_evt * mcs%integral &
            &         / real(a%n_evt_total, kind=default)
       error_cut = sqrt(a%sum_evt_sq) * mcs%integral &
            &      / real(a%n_evt_total, kind=default)
       if (integral_cut/=0) then
          rel_error_cut = error_cut / integral_cut
       else
          rel_error_cut = 0
       end if
    else
       eff = 1
       integral_cut = 0
       error_cut = 0
       rel_error_cut = 0
    end if
    accuracy = mc_get_accuracy(integral_cut, error_cut, a%n_evt)
    write(u, '(A)') "%"
10  format(2x,A,G11.4,A)
11  format(2x,A,G12.5,A,G10.3,A/,2x,A,2PF7.2,A)
    if (mcs%n_in==2) then
!        write(u, '(A)') '\begin{displaymath}'
!        write(u, 10) '\sqrt{s} = ', mcs%sqrts, '\;\textrm{GeV}\qquad'
!        write(u, 10) '\int\mathcal{L} = ', &
!             & mcs%luminosity, '\;\textrm{fb}^{-1}'
!        write(u, '(A)') '\end{displaymath}'
       write(u, '(A)') '\begin{align*}'
       write(u, 11) &
            & '\sigma_{\rm tot} &= ', mcs%integral, &
            & ' \quad \pm ', mcs%error, '\;\textrm{fb}', &
            & ' \quad [\pm ', rel_error, '\;\%]'
       write(u, *) '& n_{\rm evt,\ tot} &= ', a%n_evt_total
       write(u, *) '\\'
       write(u, 11) &
            & '\sigma_{\rm cut} &= ', integral_cut, &
            & ' \quad \pm ', error_cut, '\;\textrm{fb}', &
            & ' \quad [\pm ', rel_error_cut, '\;\%]'
       write(u, *) '& n_{\rm evt,\ cut} &= ', a%n_evt, ' \quad '
       write(u, 12) '[ ', 100*eff, ' \;\%]'
       write(u, '(A)') '\end{align*}'
    else
       write(u, '(A)') '\begin{align*}'
!       write(u, 10) 'M &=& ', mcs%sqrts, '\;\textrm{GeV}\\'
       write(u, 11) &
            & '\Gamma &=& ', mcs%integral, &
            & ' \quad \pm ', mcs%error, '\;\textrm{GeV}', &
            & ' \quad [\pm ', rel_error, '\;\%]'
       write(u, *) '& n_{\rm evt,\ tot} &= ', a%n_evt_total
       write(u, *) '\\'
       write(u, 11) &
            & '\Gamma_{\rm cut} &=& ', integral_cut, &
            & ' \quad \pm ', error_cut, '\;\textrm{GeV}', &
            & ' \quad [\pm ', rel_error_cut, '\;\%]\\'
       write(u, *) '& n_{\rm evt,\ cut} &= ', a%n_evt, ' \quad '
       write(u, 12) '[ ', 100*eff, ' \;\%]'
       write(u, '(A)') '\end{align*}'
    end if
12  format(2x,A,F7.2,A)
    write(u, '(A)') "%"
    write(u, '(A)') "%"
!     if (a%sum_evt_excess>0) then
!        call mc_write_hline(u)
!        write(u,'(1x,A,1x,F6.1,7x,A,1x,2PF7.2,1x,A)') &
!             & '! Excess events:', a%sum_evt_excess, &
!             & '(Error[%]:', a%sum_evt_excess / a%sum_evt, ')'
!     end if
!     write(u, '(A)') "\end{verbatim}"
  end subroutine gml_data_write_unit

  subroutine analysis_bundle_apply(aa, evt, weight, excess)
    type(analysis_bundle), intent(inout) :: aa
    type(event), intent(in) :: evt
    real(kind=default), intent(in), optional :: weight, excess
    integer :: i
    do i=1, aa%n_datasets
       call apply(aa%a(i), evt, weight, excess)
    end do
  end subroutine analysis_bundle_apply

  subroutine analysis_data_apply(a, evt, weight, excess)
    type(analysis_data), intent(inout) :: a
    type(event), intent(in) :: evt
    real(kind=default), intent(in), optional :: weight, excess
    logical :: ok
    integer :: i, j
    a%n_evt_total = a%n_evt_total + 1
    call apply(a%cc, evt, ok)
    if (.not.ok) return
    CHECK_HELICITY: do i=1, a%n_hel
       ok = all (a%hel(:evt%n_out,i) == evt%hel_out)
       if (ok) exit CHECK_HELICITY
    end do CHECK_HELICITY
    if (.not.ok) return
    CHECK_FLAVOR: do i=1, a%n_flv
       ok = .true.
       do j=1, evt%n_out
          if (a%flv(j,i) /= particle_code(evt%prt(j))) ok = .false.
       end do
       if (ok) exit CHECK_FLAVOR
    end do CHECK_FLAVOR
    if (.not.ok) return
    a%n_evt = a%n_evt + 1
    if (present(weight)) then
       a%sum_evt = a%sum_evt + weight
       a%sum_evt_sq = a%sum_evt_sq + weight**2
    else
       a%sum_evt = a%sum_evt + 1
       a%sum_evt_sq = a%sum_evt_sq + 1
       if (present(excess)) then
          a%sum_evt_excess = a%sum_evt_excess + excess
          if (excess > a%max_weight)  a%max_weight = 1 + excess
       end if
    end if
    do i=1, a%n_histograms
       call histogram_data_apply(a%hh(i), evt, weight, excess)
    end do
  end subroutine analysis_data_apply

  subroutine histogram_data_apply(h, evt, weight, excess)
    type(histogram_data), intent(inout) :: h
    type(event), intent(in) :: evt
    real(kind=default), intent(in), optional :: weight, excess
    real(kind=default) :: value
    type(four_momentum), dimension(2) :: p
    integer :: i
    do i=1,2
       if (h%code(i)/=0) call calculate_momentum_sum(p(i), evt, h%code(i))
    end do
    select case(h%mode)
    case(M_CUT)
       value = p(1)**1
    case(LM_CUT)
       value = abs(p(1)**1)
       if (value/=0) then
          value = log10(value)
       else
          value = -huge(1._default)
       end if
    case(MSQ_CUT)
       value = p(1)**2
    case(E_CUT)
       value = energy(p(1))
    case(LE_CUT)
       value = energy(p(1))
       if (value>0) then
          value = log10(value)
       else
          value = -huge(1._default)
       end if
    case(PT_CUT)
       value = transverse_momentum(p(1))
    case(PL_CUT)
       value = longitudinal_momentum(p(1))
    case(P_CUT)
       value = space_part(p(1))**1
    case(RAP_CUT)
       value = .5 * log( (energy(p(1)) + longitudinal_momentum(p(1))) &
            &           /(energy(p(1)) - longitudinal_momentum(p(1))))
    case(ETA_CUT)
       value = -log( tan(polar_angle(space_part(p(1))) / 2))
    case(A_CUT)
       value = angle(space_part(p(1)), space_part(p(2)))
    case(CT_CUT)
       value = angle_ct(space_part(p(1)), space_part(p(2)))
    case(AD_CUT)
       value = angle(space_part(p(1)), space_part(p(2))) &
            &  / degree
       if (value>180)  value = 180
    case(PHI_CUT)
       value = atan2(component(p(2),2), component(p(2),1)) &
          &   -atan2(component(p(1),2), component(p(1),1))
    case(PHID_CUT)
       value = ( atan2(component(p(2),2), component(p(2),1)) &
          &     -atan2(component(p(1),2), component(p(1),1))) &
          &   / degree
       if (value>180) then
          value = 180
       else if (value<-180) then
          value = -180
       end if
    case(DETA_CUT)
       value = -log( tan(polar_angle(space_part(p(2)))/2) &
            &       /tan(polar_angle(space_part(p(1)))/2))
    end select
    call fill_histogram(h%h, value, weight, excess)
  end subroutine histogram_data_apply


end module analyze
